#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <sysexits.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include "declrom.h"

#define DEBUG(fmt, ...) fprintf(stderr, "%s:%d " fmt "\n", __FILE__, __LINE__, ##__VA_ARGS__)

uint32_t sdeclrom_crc(sDeclRom *rom) {
	unsigned char *i, *inrom;
	uint32_t sum = 0;
	int n = 0;
	uint32_t len = rom->formatblock.length;

	inrom = malloc(len);
	if(!inrom) return 0;
	memcpy(inrom, rom->rom+rom->romlen-len, len);
	i = inrom;
	inrom[len-12] = 0;
	inrom[len-11] = 0;
	inrom[len-10] = 0;
	inrom[len-9] = 0;

	while(n < rom->formatblock.length) {
		uint32_t tmpint = sum >> 31;
		sum = sum << 1;
		sum |= tmpint;
		sum += *i;
		i++;
		n++;
	}

	free(inrom);
	return sum;
}

void sdeclrom_free(sDeclRom *rom) {
	if(rom->rom) free(rom->rom);
}

void sdeclrom_parse_vendorinfo(sDeclRom *rom, sResource *rsrc, uint8_t *i) {
	uint32_t entry;

	do {
		uint8_t rsrcid;
		uint32_t offset;
		entry = *(uint32_t*)i;
		entry = ntohl(entry);
		rsrcid = (entry & 0xFF000000) >> 24;
		offset = entry & 0x00FFFFFF;
		DEBUG("Found vendor info entry %d offset %d", rsrcid, offset);
		if((i+offset) > (rom->rom + rom->romlen)) {
			DEBUG("Offset is beyond the bounds of the ROM, skipping");
			break;
		}
		switch(rsrcid) {
		case 1:
			rsrc->vendorid = strdup((char*)i+offset);
			DEBUG("vendorid: %s", rsrc->vendorid);
			break;
		case 2:
			rsrc->serialnum = strdup((char*)i+offset);
			DEBUG("serialnum: %s", rsrc->serialnum);
			break;
		case 3:
			rsrc->revlevel = strdup((char*)i+offset);
			DEBUG("revlevel: %s", rsrc->revlevel);
			break;
		case 4:
			rsrc->partnum = strdup((char*)i+offset);
			DEBUG("partnum: %s", rsrc->partnum);
			break;
		case 5:
			rsrc->date = strdup((char*)i+offset);
			DEBUG("date: %s", rsrc->date);
			break;
		case 255: break;
		default:
			DEBUG("Unknown vendor entry %d", rsrcid);
			break;
		};
		i += 4;
	} while(entry != 0xFF000000);
}

sExecBlock *sdeclrom_parse_sexecblock(uint8_t *rsrc) {
	sExecBlock *ret = NULL;

	ret = calloc(1, sizeof(sExecBlock));
	ret->len = *(uint32_t*)rsrc;
	ret->len = ntohl(ret->len);
	rsrc += 4;
	ret->revision = *rsrc;
	rsrc++;
	ret->cpuid = *rsrc;
	rsrc += 3;
	ret->offset = ntohl(*(uint32_t*)rsrc);
	rsrc += 4;
	DEBUG("cpu id: %d", ret->cpuid);
	DEBUG("code length: %d", ret->len-12);
	ret->code = malloc(ret->len-12);
	memcpy(ret->code, rsrc, ret->len-12);

	return ret;
}

int sdeclrom_parse_driverdir(sResource *rsrc, uint8_t *i) {
	uint32_t entry;
	int drvrcount = 0;

	do {
		uint8_t rsrcid;
		uint32_t offset;
		entry = *(uint32_t*)i;
		entry = ntohl(entry);
		rsrcid = (entry & 0xFF000000) >> 24;
		offset = entry & 0x00FFFFFF;
		if(rsrcid == 255) break;

		uint8_t *drvr = i+offset;
		rsrc->drvrDir[drvrcount] = malloc(sizeof(sBlock));
		rsrc->drvrDir[drvrcount]->cpuid = rsrcid;
		rsrc->drvrDir[drvrcount]->len = ntohl(*(uint32_t*)drvr);
		drvr += 4;
		rsrc->drvrDir[drvrcount]->data = malloc(rsrc->drvrDir[drvrcount]->len-4);
		memcpy(rsrc->drvrDir[drvrcount]->data, drvr, rsrc->drvrDir[drvrcount]->len-4);
		drvrcount++;
		i += 4;
	} while(entry != 0xFF000000);

	return 0;
}

PixMap *sdeclrom_parse_videoparams(uint8_t *i) {
	PixMap *ret = NULL;

	i += 4; /* Skip record size, since this is fixed size anyway */

	ret = calloc(1, sizeof(PixMap));
	ret->vpBaseOffset = ntohl(*(uint32_t*)i);
	i += 4;
	ret->vpRowBytes = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpBounds[0] = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpBounds[1] = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpBounds[2] = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpBounds[3] = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpVersion = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpPackType = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpPackSize = ntohl(*(uint16_t*)i);
	i += 4;
	ret->vpHRes = ntohl(*(uint32_t*)i);
	i += 4;
	ret->vpVRes = ntohl(*(uint32_t*)i);
	i += 4;
	ret->vpPixelType = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpPixelSize = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpCmpCount = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpCmpSize = ntohs(*(uint16_t*)i);
	i += 2;
	ret->vpPlaneBytes = ntohl(*(uint32_t*)i);
	i += 4;

	return ret;
}

sRsrcVidMode *sdecl_parse_vidmode(sDeclRom *rom, sResource *rsrc, uint8_t *i) {
	uint32_t entry;
	sRsrcVidMode *new = calloc(1, sizeof(sRsrcVidMode));

	do {
		uint8_t rsrcid;
		uint32_t offset;

		entry = *(uint32_t*)i;
		entry = ntohl(entry);

		rsrcid = (entry & 0xFF000000) >> 24;
		offset = entry & 0x00FFFFFF;
		if(offset & 0x00800000) {
			offset |= 0xFF000000;
		}

		if(rsrcid == 0xff) break;

		DEBUG("Found video mode directory entry id %d with offset %x", rsrcid, offset);


		switch(rsrcid) {
		case 1: /* vid params */
			DEBUG("Found video parameters at %lx", (long)(i - rom->rom)+offset);
			new->vidParams = sdeclrom_parse_videoparams(i+offset);
			break;
		case 3: /* page count */
			new->pages = offset;
			break;
		case 4: /* dev type*/
			new->devtype = offset;
			break;
		default:
			break;
		};

		i+=4;
	} while(entry != 0xFF000000);
	return new;
}

void sdecl_parse_vidnames(sDeclRom *rom, sResource *rsrc, uint8_t *i) {
	uint32_t entry;

	rsrc->hasVidNames = 1;
	TAILQ_INIT(&rsrc->vidNames);

	do {
		uint8_t modeid;
		uint32_t offset;

		entry = *(uint32_t*)i;
		entry = ntohl(entry);

		modeid = (entry & 0xFF000000) >> 24;
		offset = entry & 0x00FFFFFF;
		if(offset & 0x00800000) {
			offset |= 0xFF000000;
		}

		if(modeid == 0xff) break;

		DEBUG("Video entry %x at %lx", modeid, (long)(i - rom->rom) +offset);
		
		sRsrcVidNameEntry *new = calloc(1, sizeof(sRsrcVidNameEntry));
		new->reclen = ntohl(*(uint32_t*)(i+offset));
		new->resid = ntohs(*(uint16_t*)(i+offset+4));
		new->name = calloc(1, new->reclen);
		new->modeid = modeid;
		memcpy(new->name, i+offset+6, new->reclen-6);
		DEBUG("Video entry: %x %x %s", new->reclen, new->resid, new->name);
		TAILQ_INSERT_TAIL(&rsrc->vidNames, new, next);

		i+=4;
	} while(entry != 0xFF000000);
}

sResource *sdeclrom_parse_sresource(sDeclRom *rom, uint8_t *rsrc) {
	sResource *ret = NULL;
	uint32_t entry;

	ret = calloc(1, sizeof(sResource));
	if(!ret) return ret;

	TAILQ_INIT(&ret->vidModes);

	do {
		uint8_t rsrcid;
		int32_t offset;

		entry = *(uint32_t*)rsrc;
		entry = ntohl(entry);
		rsrcid = (entry & 0xFF000000) >> 24;
		offset = entry & 0x00FFFFFF;
		if(offset & 0x00800000) {
			offset |= 0xFF000000;
		}

		DEBUG("Found resource entry %d offset %d %x", rsrcid, offset, offset);
		switch(rsrcid) {
		case 1: /* type */
			ret->type = calloc(1, sizeof(sRsrcType));
			ret->type->category = ntohs(*(uint16_t*)&rsrc[offset]);
			ret->type->cType = ntohs(*(uint16_t*)&rsrc[offset+2]);
			ret->type->drSW = ntohs(*(uint16_t*)&rsrc[offset+4]);
			ret->type->drHW = ntohs(*(uint16_t*)&rsrc[offset+6]);
			DEBUG("category %x type %x sw %x hw %x", ret->type->category, ret->type->cType, ret->type->drSW, ret->type->drHW);
			break;
		case 2: /* name */
			ret->name = strdup((char*)(rsrc+offset));
			DEBUG("name: %s", ret->name);
			break;
		case 4: /* driver directory */
			DEBUG("Found driver directory at %d %x", offset, offset);
			sdeclrom_parse_driverdir(ret, rsrc+offset);
			break;
		case 6: /* bootrec */
			DEBUG("Found bootrec at %d %x", offset, offset);
			ret->bootRec = sdeclrom_parse_sexecblock(rsrc+offset);
			break;
		case 7: /* flags */
			DEBUG("Found flags of: %x", offset);
			ret->flags = entry;
			break;
		case 8: /* sRsrcHWDevID */
			DEBUG("Found devid of: %x", offset);
			ret->devid = entry;
			break;
		case 10: /* minorbaseos */
			ret->minBaseOS = malloc(4);
			*ret->minBaseOS = ntohl(*(uint32_t*)&rsrc[offset]);
			DEBUG("minBaseOS: %x", *ret->minBaseOS);
			break;
		case 11: /* minorLength */
			ret->minLen = malloc(4);
			*ret->minLen = ntohl(*(uint32_t*)&rsrc[offset]);
			DEBUG("minLen: %x", *ret->minLen);
			break;
		case 32: /* board id */
			ret->boardid = offset & 0xFF;
			DEBUG("boardid: %x", ret->boardid);
			break;
		case 34: /* primaryInit */
			DEBUG("Found primary Init at %d %x", offset, offset);
			ret->primaryInit = sdeclrom_parse_sexecblock(rsrc+offset);
			break;
		case 36: /* vendorinfo */
			if((rsrc+offset) > (rom->rom + rom->romlen)) {
				DEBUG("Vendor offset is beyond end of ROM, skipping");
			}
			sdeclrom_parse_vendorinfo(rom, ret, rsrc+offset);
			break;
		case 38: /* secondaryInit */
			DEBUG("Found secondary Init at %d %x", offset, offset);
			ret->secondaryInit = sdeclrom_parse_sexecblock(rsrc+offset);
			break;
		case 65: /* sRsrcVidNames */
			DEBUG("Found video mode names directory at %d %x", offset, offset);
			sdecl_parse_vidnames(rom, ret, rsrc+offset);
			break;
		case 255: break;
		default:
			if((ret->type->category == catDisplay) && (ret->type->drSW == 1) && (rsrcid >= 128)) {
				DEBUG("Found video mode entry %d at %lx", rsrcid, (long)(rom->rom - rsrc) + offset);
				sRsrcVidMode *vmode = sdecl_parse_vidmode(rom, ret, rsrc+offset);
				vmode->modeid = rsrcid;
				TAILQ_INSERT_TAIL(&ret->vidModes, vmode, next);
			} else {
				DEBUG("Unknown resource entry %d ", rsrcid);
			}
			// Sooo dangerous, this only works if the unknown is
			// a flag instead of an offset
#if 0
			ret->unknowncount++;
			ret->unknowns = realloc(ret->unknowns, ret->unknowncount);
			ret->unknowns[ret->unknowncount-1] = entry;
#endif
			break;
		};
		rsrc += 4;
	} while(entry != 0xFF000000);

	return ret;
}

int sdeclrom_parse_resources(sDeclRom *rom) {
	uint32_t entry = 0;
	uint8_t *i = rom->rom + rom->romlen - 20 + rom->formatblock.diroffset;

	do {
		uint8_t rsrcid;
		uint32_t offset;
		sResource *resource;

		entry = *(uint32_t*)i;
		entry = ntohl(entry);
		rsrcid = (entry & 0xFF000000) >> 24;
		offset = entry & 0x00FFFFFF;
		DEBUG("Found entry %d offset %d", rsrcid, offset);
		resource = sdeclrom_parse_sresource(rom, i+offset);
		if(resource) {
			resource->rsrcid = rsrcid;
			TAILQ_INSERT_TAIL(&rom->resourcelist, resource, next);
		}
		i += 4;
	} while(entry != 0xFF000000);

	return 0;
}

sDeclRom *sdeclrom_readfile(char *infile) {
	sDeclRom *ret;
	int ifd;
	struct stat sb;
	int err = -1;

	ret = calloc(1, sizeof(sDeclRom));
	if(!ret) {
		DEBUG("Failed to allocate sDeclRom");
		goto out;
	}

	ifd = open(infile, O_RDONLY);
	if(ifd < 0) {
		DEBUG("Failed to open %s: %s", infile, strerror(errno));
		goto out;
	}

	if(fstat(ifd, &sb) != 0) {
		DEBUG("Failed to fstat %s: %s", infile, strerror(errno));
		goto out;
	}

	ret->romlen = (uint32_t)sb.st_size;
	ret->rom = malloc(ret->romlen);
	if(!ret->rom) goto out;

	if(read(ifd, ret->rom, ret->romlen) != ret->romlen) {
		DEBUG("Failed to read %s: %s", infile, strerror(errno));
		goto out;
	}

	ret->formatblock.bytelanes = ret->rom[ret->romlen-1];
	ret->formatblock.testpattern = *(uint32_t*)(&ret->rom[ret->romlen-6]);
	ret->formatblock.testpattern = ntohl(ret->formatblock.testpattern);
	if(ret->formatblock.testpattern != kDeclRomPattern) {
		DEBUG("format block test pattern doesn't match: %x", ret->formatblock.testpattern);
		goto out;
	}
	ret->formatblock.format = ret->rom[ret->romlen-7];
	ret->formatblock.revision = ret->rom[ret->romlen-8];
	ret->formatblock.crc = *(uint32_t*)(&ret->rom[ret->romlen-12]);
	ret->formatblock.crc = ntohl(ret->formatblock.crc);
	ret->formatblock.length = *(uint32_t*)(&ret->rom[ret->romlen-16]);
	ret->formatblock.length = ntohl(ret->formatblock.length);
	ret->formatblock.diroffset = *(uint32_t*)(&ret->rom[ret->romlen-20]);
	ret->formatblock.diroffset = 0xFF000000 | ntohl(ret->formatblock.diroffset);

	uint32_t computedcrc = sdeclrom_crc(ret);
	if(computedcrc != ret->formatblock.crc) {
		DEBUG("CRCs don't match.  computed: %x, found: %x", computedcrc, ret->formatblock.crc);
	}

	TAILQ_INIT(&ret->resourcelist);
	sdeclrom_parse_resources(ret);

	err = 0;
out:
	if(err) {
	}
	return ret;
}

void sdeclrom_growheap(sDeclRom *rom, uint32_t increment, uint8_t **table, uint8_t **heap) {
	uint32_t tableoff = 0;
	uint32_t heapoff = 0;
	if(table) {
		tableoff = (uint32_t)(*table - rom->rom);
	}
	if(heap) {
		heapoff = (uint32_t)(*heap - rom->rom);
	}
	rom->rom = realloc(rom->rom, rom->romlen + increment);
	rom->romlen += increment;
	if(table) {
		*table =  rom->rom + tableoff;
	}
	if(heap) {
		*heap =  rom->rom + heapoff;
	}
}

void sdeclrom_sexec_quiesce(sDeclRom *rom, sExecBlock *block, uint8_t **table, uint8_t **heap) {
	sdeclrom_growheap(rom, block->len, table, heap);
	*(uint32_t*)*heap = htonl(block->len);
	*heap += 4;
	*(uint32_t*)*heap = htonl( (block->revision << 24) | (block->cpuid << 16) );
	*heap += 4;
	*(uint32_t*)*heap = htonl(block->offset);
	*heap += 4;
	memcpy(*heap, block->code, block->len - 12);
	*heap += block->len - 12;
	return;
}

int sdeclrom_rsrc_quiesce(sDeclRom *rom, sResource *rsrc, uint8_t *heap) {
	int ret = -1;
	uint32_t tablesize = 0;
	uint8_t *table = heap;

	if(!rsrc->type) {
		DEBUG("Resource %d lacks required type, skipping", rsrc->rsrcid);
		goto out;
	}
	if(!rsrc->name) {
		DEBUG("Resource lacks required name, skipping");
		goto out;
	}

	tablesize += 4; /* type */
	tablesize += 4; /* name */
	if(rsrc->drvrDir[0]) {
		tablesize += 4;
	}
	if(rsrc->loadRec) {
		tablesize += 4;
	}
	if(rsrc->bootRec) {
		tablesize += 4;
	}
	if(rsrc->flags) {
		tablesize += 4;
	}
	if(rsrc->devid) {
		tablesize += 4;
	}
	if(rsrc->primaryInit) {
		tablesize += 4;
	}
	if(rsrc->minBaseOS) {
		tablesize += 4;
	}
	if(rsrc->minLen) {
		tablesize += 4;
	}
	if(rsrc->boardid) {
		tablesize += 4;
	}
	if(rsrc->vendorid || rsrc->serialnum || rsrc->revlevel || rsrc->partnum || rsrc->date) {
		tablesize += 4;
	}
	if(rsrc->secondaryInit) {
		tablesize += 4;
	}
	if(rsrc->unknowncount) {
		tablesize += rsrc->unknowncount*4;
	}
	tablesize += 4; /* EOL */
	sdeclrom_growheap(rom, tablesize, &table, &heap);
	heap += tablesize;
	
	sdeclrom_growheap(rom, 8, &table, &heap);
	heap = table + tablesize;
	*(uint32_t*)table = htonl( (1 << 24) | ((heap - table) & 0x00FFFFFF) );
	table += 4;

	DEBUG("writing category: %x %d", rsrc->type->category, rsrc->type->category);
	*(uint16_t*)heap = htons(rsrc->type->category);
	heap += 2;
	*(uint16_t*)heap = htons(rsrc->type->cType);
	heap += 2;
	*(uint16_t*)heap = htons(rsrc->type->drSW);
	heap += 2;
	*(uint16_t*)heap = htons(rsrc->type->drHW);
	heap += 2;

	sdeclrom_growheap(rom, strlen(rsrc->name)+1, &table, &heap);
	*(uint32_t*)table = htonl( (2 << 24) | ((heap - table) & 0x00FFFFFF) );
	table += 4;
	memcpy(heap, rsrc->name, strlen(rsrc->name)+1);
	heap += strlen(rsrc->name)+1;

	if(rsrc->drvrDir[0]) {
		int i;
		int drvrcount = 0;

		*(uint32_t*)table = htonl( (4 << 24) | ((heap - table) & 0x00FFFFFF) );
		table += 4;

		uint32_t tableoff = (uint32_t)(table - rom->rom);

		table = heap;
		for(drvrcount = 0; rsrc->drvrDir[drvrcount]; drvrcount++);
		sdeclrom_growheap(rom, (drvrcount+1) * 4, &table, &heap);
		heap += (drvrcount+1) * 4;
		
		for(i = 0; i < drvrcount; i++) {
			if(rsrc->drvrDir[i]->nowrite) {
				*(uint32_t*)table = htonl( (rsrc->drvrDir[i]->cpuid << 24) | (rsrc->drvrDir[i]->nowrite & 0x00FFFFFF) );
				table += 4;
				continue;
			}
			*(uint32_t*)table = htonl( (rsrc->drvrDir[i]->cpuid << 24) | ((heap - table) & 0x00FFFFFF) );
			table += 4;
			sdeclrom_growheap(rom, rsrc->drvrDir[i]->len, &table, &heap);
			*(uint32_t*)heap = htonl(rsrc->drvrDir[i]->len);
			heap += 4;
			memcpy(heap, rsrc->drvrDir[i]->data, rsrc->drvrDir[i]->len-4);
			heap += rsrc->drvrDir[i]->len-4;
		}
		*(uint32_t*)table = htonl((255<<24) & 0xFF000000);

		table = rom->rom + tableoff;
	}

	if(rsrc->loadRec) {
		*(uint32_t*)table = htonl( (5 << 24) | ((heap - table) & 0x00FFFFFF) );
		table += 4;
		sdeclrom_sexec_quiesce(rom, rsrc->loadRec, &table, &heap);
	}

	if(rsrc->bootRec) {
		if(rsrc->nowriteBootRec) {
			*(uint32_t*)table = htonl( (6 << 24) | (rsrc->nowriteBootRec & 0x00FFFFFF) );
			table += 4;
		} else {
			*(uint32_t*)table = htonl( (6 << 24) | ((heap - table) & 0x00FFFFFF) );
			table += 4;
			sdeclrom_sexec_quiesce(rom, rsrc->bootRec, &table, &heap);
		}
	}

	if(rsrc->flags) {
		*(uint32_t*)table = htonl( (7 << 24) | (rsrc->flags & 0x00FFFFFF) );
		table += 4;
	}

	if(rsrc->devid) {
		*(uint32_t*)table = htonl( (8 << 24) | (rsrc->devid & 0x00FFFFFF) );
		table += 4;
	}

	if(rsrc->minBaseOS) {
		*(uint32_t*)table = htonl( (10 << 24) | ((heap - table) & 0x00FFFFFF) );
		table += 4;
		sdeclrom_growheap(rom, 4, &table, &heap);
		*(uint32_t*)heap = htonl(*rsrc->minBaseOS);
		heap += 4;
	}

	if(rsrc->minLen) {
		*(uint32_t*)table = htonl( (11 << 24) | ((heap - table) & 0x00FFFFFF) );
		table += 4;
		sdeclrom_growheap(rom, 4, &table, &heap);
		*(uint32_t*)heap = htonl(*rsrc->minLen);
		heap += 4;
	}

	if(rsrc->boardid) {
		*(uint32_t*)table = htonl( (32 << 24) | rsrc->boardid );
		table += 4;
	}

	if(rsrc->primaryInit) {
		*(uint32_t*)table = htonl( (34 << 24) | ((heap - table) & 0x00FFFFFF) );
		table += 4;
		sdeclrom_sexec_quiesce(rom, rsrc->primaryInit, &table, &heap);
	}

	if(rsrc->vendorid || rsrc->serialnum || rsrc->revlevel || rsrc->partnum || rsrc->date) {
		*(uint32_t*)table = htonl( (36 << 24) | ((heap - table) & 0x00FFFFFF) );
		table += 4;

		uint32_t tableoff = (uint32_t)(table - rom->rom);
		table = heap;
		uint32_t tablesize = 0;

		if(rsrc->vendorid) tablesize += 4;
		if(rsrc->serialnum) tablesize += 4;
		if(rsrc->revlevel) tablesize += 4;
		if(rsrc->partnum) tablesize += 4;
		if(rsrc->date) tablesize += 4;
		tablesize += 4; /*EOL*/
		sdeclrom_growheap(rom, tablesize, &table, &heap);
		heap += tablesize;
	
		if(rsrc->vendorid) {
			*(uint32_t*)table = htonl( (1 << 24) | ((heap - table) & 0x00FFFFFF) );
			table += 4;
			sdeclrom_growheap(rom, strlen(rsrc->vendorid)+1, &table, &heap);
			memcpy(heap, rsrc->vendorid, strlen(rsrc->vendorid)+1);
			heap += strlen(rsrc->vendorid)+1;
		}
		if(rsrc->serialnum) {
			*(uint32_t*)table = htonl( (2 << 24) | ((heap - table) & 0x00FFFFFF) );
			table += 4;
			sdeclrom_growheap(rom, strlen(rsrc->serialnum)+1, &table, &heap);
			memcpy(heap, rsrc->serialnum, strlen(rsrc->serialnum)+1);
			heap += strlen(rsrc->serialnum)+1;
		}
		if(rsrc->revlevel) {
			*(uint32_t*)table = htonl( (3 << 24) | ((heap - table) & 0x00FFFFFF) );
			table += 4;
			sdeclrom_growheap(rom, strlen(rsrc->revlevel)+1, &table, &heap);
			memcpy(heap, rsrc->revlevel, strlen(rsrc->revlevel)+1);
			heap += strlen(rsrc->revlevel)+1;
		}
		if(rsrc->partnum) {
			*(uint32_t*)table = htonl( (4 << 24) | ((heap - table) & 0x00FFFFFF) );
			table += 4;
			sdeclrom_growheap(rom, strlen(rsrc->partnum)+1, &table, &heap);
			memcpy(heap, rsrc->partnum, strlen(rsrc->partnum)+1);
			heap += strlen(rsrc->partnum)+1;
		}
		if(rsrc->date) {
			*(uint32_t*)table = htonl( (5 << 24) | ((heap - table) & 0x00FFFFFF) );
			table += 4;
			sdeclrom_growheap(rom, strlen(rsrc->date)+1, &table, &heap);
			memcpy(heap, rsrc->date, strlen(rsrc->date)+1);
			heap += strlen(rsrc->date)+1;
		}
		*(uint32_t*)table = htonl( (255 << 24) & 0xFF000000 );
		table = rom->rom + tableoff;
	}

	if(rsrc->secondaryInit) {
		*(uint32_t*)table = htonl( (38 << 24) | ((heap - table) & 0x00FFFFFF) );
		table += 4;
		sdeclrom_sexec_quiesce(rom, rsrc->secondaryInit, &table, &heap);
	}

#if 0
	if(rsrc->unknowncount) {
		int i;
		for(i = 0; i < rsrc->unknowncount; i++) {
			*(uint32_t*)table = htonl(rsrc->unknowns[i]);
			table += 4;
		}
	}
#endif

	*(uint32_t*)table = htonl( (255 << 24) & 0xFF000000 );

	ret = 0;
out:
	return ret;
}

int sdeclrom_quiesce(sDeclRom *rom) {
	int ret = -1;
	uint32_t numrsrcs = 0;
	sResource *i = NULL;
	uint8_t *direntry = NULL;
	uint8_t *heap = NULL;
	uint32_t diroffset = 0;
	uint8_t *crc;

	if(rom->rom) free(rom->rom);
	rom->romlen = 0;
	rom->rom = NULL;

	TAILQ_FOREACH(i, &rom->resourcelist, next) {
		numrsrcs++;
	}
	
	/* directory list will be 32bits * number of entries + end of list */
	rom->romlen = (numrsrcs) * 4;
	rom->rom = calloc(1,rom->romlen);

	TAILQ_FOREACH(i, &rom->resourcelist, next) {
		direntry = rom->rom +diroffset;
		if(i->rsrcid == 255) {
			*(uint32_t*)direntry = htonl((255 << 24) & 0xFF000000);
			break;
		} else {
			*(uint32_t*)direntry = htonl((i->rsrcid << 24) | (uint32_t)((rom->rom +rom->romlen) - direntry));
			sdeclrom_rsrc_quiesce(rom, i, rom->rom + rom->romlen);
			diroffset += 4;
		}
	}

	heap = rom->rom + rom->romlen;
	sdeclrom_growheap(rom, 20, NULL, &heap);
	int32_t offset = (int32_t)(heap - rom->rom);
	offset = 0 - offset;
	*(uint32_t*)heap = htonl( offset & 0x00FFFFFF );
	heap += 4;
	rom->formatblock.length = rom->romlen;
	*(uint32_t*)heap = htonl( rom->romlen );
	heap += 4;
	crc = heap;
	*(uint32_t*)heap = 0; /* leave CRC blank until the end */
	heap += 4;
	*heap = rom->formatblock.revision;
	heap++;
	*heap = rom->formatblock.format;
	heap++;
	*(uint32_t*)heap = htonl( kDeclRomPattern );
	heap += 4;
	*heap = 0;
	heap++;
	*heap = rom->formatblock.bytelanes;;
	heap++;

	*(uint32_t*)crc = htonl( sdeclrom_crc(rom) );

	ret = 0;
	return ret;
}

int sdeclrom_writefile(sDeclRom *rom, char *file) {
	int ofd = -1;
	int ret = -1;

	sdeclrom_quiesce(rom);
	errno = 0;
	ofd = open(file, O_WRONLY | O_TRUNC | O_CREAT, 0644);
	if(ofd < 0) {
		DEBUG("opening of output file failed: %d %s", errno, strerror(errno));
		goto out;
	}

	errno = 0;
	if(write(ofd, rom->rom, rom->romlen) != rom->romlen) {
		DEBUG("writing of rom failed: %d %s", errno, strerror(errno));
		goto out;
	}

out:
	close(ofd);
	return ret;
}
